#![no_std]
#![no_main]

use ble::{gatt::Server, msg_service::MAX_STRING_LENGTH};
use defmt_rtt as _;
use embassy_nrf::{self as _, gpio::Pin};
use embassy_time::Delay;
use embedded_hal_async::delay::DelayNs;
use futures::pin_mut;
use nrf_softdevice::{Flash, Softdevice};
use panic_probe as _;

use crate::lcd::Lcd;
use defmt::{error, info, unwrap};
use embassy_executor::Spawner;

pub mod events;
pub mod ble {
    pub mod advertisment;
    pub mod gatt;
    pub mod msg_service;
}
pub mod flash;
pub mod lcd;
pub mod peripherals;
pub mod softdevice;

const DEVICE_NAME: &str = "alert-me display";

#[embassy_executor::task]
async fn ble(sd: &'static Softdevice, server: Server) {
    loop {
        let conn = ble::advertisment::run(sd, DEVICE_NAME).await;
        let error = ble::gatt::run(&conn, &server).await;
        info!("gatt_server run exited with error: {:?}", error);
    }
}

#[embassy_executor::task]
async fn lcd_scroll(sd: &'static Softdevice, mut lcd: lcd::Lcd<'static>, delay: u32) {
    fn until_first_zero(data: &[u8]) -> &[u8] {
        match data.iter().position(|&x| x == 0) {
            Some(pos) => &data[..pos],
            None => data,
        }
    }

    let mut d = Delay {};
    let mut event_sub = events::EVENTS.subscriber().unwrap();

    loop {
        if let Some(events::Event::Message(new_msg)) = event_sub.try_next_message_pure() {
            let new_msg = until_first_zero(&new_msg);

            match core::str::from_utf8(new_msg) {
                Ok(msg) => lcd.write(msg, lcd::Lines::First).await.unwrap(),
                Err(_) => error!("failed to parse incoming message as utf8"),
            }
        }

        lcd.scroll_buffer()
            .await
            .unwrap_or_else(|e| error!("Failed to scroll lcd: {}", e));

        d.delay_ms(delay).await;
    }
}

#[embassy_executor::task]
async fn config(sd: &'static Softdevice) {
    let flash = Flash::take(sd);
    pin_mut!(flash);

    let config = flash::Config::new(flash);

    loop {}
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let peripherals = peripherals::init();
    let mut lcd = Lcd::new(
        0x27,
        peripherals.TWISPI0,
        peripherals.P0_03.degrade(),
        peripherals.P0_04.degrade(),
    );
    lcd.init().await.expect("Failed to initialize lcd");
    let (sd, server) = softdevice::run(spawner, DEVICE_NAME).expect("Failed to run softdevice");

    unwrap!(spawner.spawn(ble(sd, server)));
    unwrap!(spawner.spawn(lcd_scroll(sd, lcd, 500)));
    unwrap!(spawner.spawn(config(sd)));
}
